# python-mtga-helper
# Copyright 2025 Lubosz Sarnecki <lubosz@gmail.com>
# SPDX-License-Identifier: MIT

# Heavily inspired by limited-grades by Youssef Moussaoui
# See https://github.com/youssefm/limited-grades

from enum import StrEnum
import colorsys

from scipy.stats import norm
from termcolor import colored
import numpy as np

class Grade(StrEnum):
    A_PLUS = "A+"
    A = "A"
    A_MINUS = "A-"
    B_PLUS = "B+"
    B = "B"
    B_MINUS = "B-"
    C_PLUS = "C+"
    C = "C"
    C_MINUS = "C-"
    D_PLUS = "D+"
    D = "D"
    D_MINUS = "D-"
    F = "F"

GRADE_THRESHOLDS = {
    Grade.A_PLUS: 99,
    Grade.A: 95,
    Grade.A_MINUS: 90,
    Grade.B_PLUS: 85,
    Grade.B: 76,
    Grade.B_MINUS: 68,
    Grade.C_PLUS: 57,
    Grade.C: 45,
    Grade.C_MINUS: 36,
    Grade.D_PLUS: 27,
    Grade.D: 17,
    Grade.D_MINUS: 5,
    Grade.F: 0,
}

def grade_to_color(grade: Grade) -> tuple[int, int, int]:
    threshold: int = GRADE_THRESHOLDS[grade]
    hue = threshold / (3 * 100.0)
    rgb_float = colorsys.hsv_to_rgb(hue, 1.0, 1.0)
    rgb_int = [int(c * 255) for c in rgb_float]
    return tuple[int, int, int](rgb_int)

def score_to_grade(score: float):
    for grade, threshold in GRADE_THRESHOLDS.items():
        if score >= threshold:
            return grade
    return Grade.F

def score_to_grade_string(score: float) -> str:
    if not score:
        return ""
    grade = score_to_grade(score)
    color = grade_to_color(grade)
    return colored(str(grade), color=color)

def get_mean_and_std_dev(rankings, key) -> tuple[float, float]:
    win_rates = []
    for card in rankings:
        if card[key]:
            win_rates.append(card[key])

    mean = np.mean(win_rates)
    std = np.std(win_rates, ddof=1)

    return float(mean), float(std)

def calculate_grade_scores(rankings_by_arena_id, set_rankings):
    winrates_mean, win_rates_std = get_mean_and_std_dev(set_rankings, "ever_drawn_win_rate")

    for arena_id, ranking in rankings_by_arena_id.items():
        ranking["ever_drawn_score"] = None
        if ranking["ever_drawn_win_rate"]:
            cdf = norm.cdf(ranking["ever_drawn_win_rate"], loc=winrates_mean, scale=win_rates_std)
            ranking["ever_drawn_score"] = cdf * 100

    return rankings_by_arena_id



# python-mtga-helper
# Copyright 2025 Lubosz Sarnecki <lubosz@gmail.com>
# SPDX-License-Identifier: MIT

import argparse
import logging
from pathlib import Path

import coloredlogs

from mtga_helper.limited import print_limited_course_info, bot_draft_pick_cb, premier_draft_pick_cb
from mtga_helper.mtga_log import get_log_path, get_limited_courses, follow_player_log, print_courses

logger = logging.getLogger(__name__)

def got_courses_cb(event: dict, args: argparse.Namespace):
    courses = event["Courses"]

    if args.verbose:
        print_courses(courses)

    sealed_courses = get_limited_courses(courses)
    logger.info(f"Found {len(sealed_courses)} ongoing limited games.")
    for course in sealed_courses:
        print_limited_course_info(course, args)

def business_events_cb(event: dict, args):
    if "DraftId" in event:
        premier_draft_pick_cb(event, args)

def main():
    parser = argparse.ArgumentParser(prog='mtga-helper',
                                     description='Analyse MTGA log for sealed pools with 17lands data.',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-l','--log-path', type=Path, help="Custom Player.log path")
    parser.add_argument('--land-count', type=int, help="Target Land count", default=17)
    parser.add_argument('--print-top-pairs', type=int, help="Top color pairs to print", default=3)
    parser.add_argument('-v', '--verbose', help="Log some intermediate steps", action="store_true")
    parser.add_argument('-d', '--data-set', choices=['PremierDraft', 'TradDraft', 'Sealed', 'TradSealed'],
                        help="Use specific 17lands format data set", default="PremierDraft")
    args = parser.parse_args()

    field_styles = coloredlogs.DEFAULT_FIELD_STYLES
    field_styles["levelname"]["color"] = "white"
    field_styles["funcName"] = {'color': 'blue'}
    level_styles = coloredlogs.DEFAULT_LEVEL_STYLES
    level_styles["debug"] = {'color': 'magenta', 'faint': True}
    log_level = logging.DEBUG if args.verbose else logging.INFO
    coloredlogs.install(fmt="%(levelname)s %(funcName)s %(message)s",
                        field_styles=field_styles, level=log_level, level_styles=level_styles)

    if args.log_path:
        player_log_path = args.log_path
        if not player_log_path.exists():
            logger.error(f"Can't find log file at {player_log_path}")
            return
    else:
        try:
            player_log_path = get_log_path()
        except RuntimeError:
            logger.error("Could not find MTGA log file")
            return

    try:
        start_callbacks = {
            "LogBusinessEvents": business_events_cb,
        }
        end_callbacks = {
            "EventGetCoursesV2": got_courses_cb,
            "BotDraftDraftStatus": bot_draft_pick_cb,
            "BotDraftDraftPick": bot_draft_pick_cb,
        }
        follow_player_log(player_log_path, args, start_callbacks, end_callbacks)
    except KeyboardInterrupt:
        logger.debug("Bye")


if __name__ == "__main__":
    main()


# python-mtga-helper
# Copyright 2025 Lubosz Sarnecki <lubosz@gmail.com>
# SPDX-License-Identifier: MIT

import argparse
import json
import logging
import os
import re
import time
from datetime import datetime
from io import TextIOWrapper
from pathlib import Path
from typing import Iterator

from tabulate import tabulate

logger = logging.getLogger(__name__)

MTGA_STEAM_APP_ID = 2141910

def follow(file: TextIOWrapper) -> Iterator[str]:
    current_inode: int = os.fstat(file.fileno()).st_ino

    while True:
        line = file.readline()
        if not line:
            # Handle file recreation
            inode = os.stat(file.name).st_ino
            if inode != current_inode:
                logger.info("Log file recreated")
                file.close()
                file = open(file.name, "r")
                current_inode = inode
                continue

            time.sleep(0.1)
            continue
        yield line.strip()

def get_log_path() -> Path:
    steam_path = Path.home() / ".local/share/Steam"
    if not steam_path.exists():
        raise RuntimeError("Could not find user steam path.")

    mtga_compatibility_data_path = steam_path / f"steamapps/compatdata/{MTGA_STEAM_APP_ID}"
    if not mtga_compatibility_data_path.exists():
        raise RuntimeError("Could not find MTGA compat data path.")

    prefix_c_path = mtga_compatibility_data_path / "pfx/drive_c"
    if not prefix_c_path.exists():
        raise RuntimeError("Could not find proton prefix C path.")

    PREFIX_USER_NAME = "steamuser"
    mtga_app_data_path = prefix_c_path / f"users/{PREFIX_USER_NAME}/AppData/LocalLow/Wizards Of The Coast/MTGA"
    if not mtga_app_data_path.exists():
        raise RuntimeError("Could not find MTGA user data path.")

    player_log_path = mtga_app_data_path / "Player.log"
    if not player_log_path.exists():
        raise RuntimeError("Could not find player log.")

    logger.info(f"Found MTGA log at {player_log_path}")

    return player_log_path

SUPPORTED_LIMITED_FORMATS = {
    "QuickDraft", "PremierDraft", "Sealed"
}

def get_limited_courses(courses: list) -> list:
    limited_courses = []
    for course in courses:
        if course["CardPool"]:
            for supported_format in SUPPORTED_LIMITED_FORMATS:
                if course["InternalEventName"].startswith(supported_format):
                    limited_courses.append(course)
    return limited_courses

def follow_player_log(player_log_path: Path, args: argparse.Namespace, start_callbacks, end_callbacks):
    with player_log_path.open('r') as player_log_file:
        next_line_event = ""
        for line in follow(player_log_file):
            if next_line_event:
                if next_line_event in end_callbacks:
                    try:
                        payload = json.loads(line)
                    except json.decoder.JSONDecodeError:
                        # In case of LogBusinessEvents the payload is in the start line
                        # and the following line only returns a status string
                        payload = line
                    end_callbacks[next_line_event](payload, args)
                else:
                    logger.debug(f"Unhandled end line event {next_line_event}")
                next_line_event = ""

            elif "Version:" in line and line.count("/") == 2:
                mtga_version = line.split("/")[1].strip()
                logger.info(f"Found game version {mtga_version}")
            elif "DETAILED LOGS" in line:
                detailed_log_status = line.split(":")[1].strip()
                if detailed_log_status == "DISABLED":
                    logger.warning("Detailed logs are disabled!")
                    logger.warning("Enable `Options -> Account -> Detailed Logs (Plugin Support)`")
                else:
                    logger.info(f"Detailed logs are {detailed_log_status}!")

            # Find json lines
            elif line.startswith("<=="):
                match = re.search(r"<== (\w+)\(([a-f0-9-]+)\)", line)
                if match:
                    next_line_event = match.group(1)
                    # next_line_event_id = match.group(2)

            # Find json in start line
            elif line.startswith("[UnityCrossThreadLogger]==>"):
                match = re.search(r"\[UnityCrossThreadLogger\]==> (\w+) (.*)", line)
                if match:
                    current_line_event = match.group(1)
                    outer_json = match.group(2)
                    outer_json_json_data = json.loads(outer_json)
                    inner_json_data = json.loads(outer_json_json_data["request"])
                    if current_line_event in start_callbacks:
                        start_callbacks[current_line_event](inner_json_data, args)
                    else:
                        logger.debug(f"Unhandled start line event {current_line_event}")

def time_str_to_dt(time_str: str) -> datetime:
    time_str = time_str.strip('"')
    parts = time_str.split('.')
    # shorten microseconds to 6 digits
    if len(parts[1]) > 12:
        pre_tz_str = parts[0] + '.' + parts[1][:6]
        tz_str = parts[1][7:]
        time_str = pre_tz_str + tz_str
    return datetime.strptime(time_str, "%Y-%m-%dT%H:%M:%S.%f%z")

def print_courses(courses: list):
    table = []

    print()
    print("== Courses ==")
    print()

    for course in courses:
        wins = "N/A"
        if "CurrentWins" in course:
            wins = course["CurrentWins"]

        losses = "N/A"
        if "CurrentLosses" in course:
            losses = course["CurrentLosses"]

        summary = course["CourseDeckSummary"]

        deck_name = "N/A"
        if "Name" in summary:
            deck_name = summary["Name"]

        attribs = {}
        for attrib in summary["Attributes"]:
            k = attrib["name"]
            v = attrib["value"]
            attribs[k] = v

        event_format = "N/A"
        last_updated = "N/A"
        if attribs:
            event_format = attribs["Format"]
            if "LastUpdated" in attribs:
                last_updated_dt = time_str_to_dt(attribs["LastUpdated"])
                last_updated = last_updated_dt.date().isoformat()

        deck_name = deck_name.replace("?=?Loc/Decks/Precon/", "")

        row = (
            deck_name,
            course["InternalEventName"],
            event_format,
            len(course["CardPool"]),
            wins, losses,
            last_updated,
        )
        table.append(row)

    print(tabulate(table, headers=(
        "Deck Name",
        "Event",
        "Format",
        "Pool",
        "Wins",
        "Losses",
        "Updated"
    )))



# python-mtga-helper
# Copyright 2025 Lubosz Sarnecki <lubosz@gmail.com>
# SPDX-License-Identifier: MIT

LIMITED_DECK_SIZE = 40

COLOR_PAIRS = {
    # Allied
    "WU": "Azorius",
    "UB": "Dimir",
    "BR": "Rakdos",
    "RG": "Gruul",
    "GW": "Selesnya",
    # Enemy
    "WB": "Orzhov",
    "UR": "Izzet",
    "BG": "Golgari",
    "RW": "Boros",
    "GU": "Simic"
}

def color_id_to_emoji(color_id: str):
    match color_id:
        case "W":
            return "⚪"
        case "B":
            return "⚫"
        case "U":
            return "🔵"
        case "R":
            return "🔴"
        case "G":
            return "🟢"
        case _:
            return ""

def rarity_to_emoji(rarity: str):
    match rarity:
        case "common":
            return "⬛"
        case "uncommon":
            return "⬜"
        case "rare":
            return "🟨"
        case "mythic":
            return "🟥"
        case _:
            return ""

def format_color_id_emoji(colors: str):
    return "".join(color_id_to_emoji(c) for c in colors)

def land_string_to_colors(land_type_str: str):
    found_colors = set()
    for chunk in land_type_str.split():
        match chunk:
            case "Plains":
                found_colors.add("W")
            case "Island":
                found_colors.add("U")
            case "Swamp":
                found_colors.add("B")
            case "Mountain":
                found_colors.add("R")
            case "Forest":
                found_colors.add("G")

    return "".join(list(found_colors))

def are_card_colors_in_pair(card_colors: str, color_pair: str) -> bool:
    for card_color in card_colors:
        if card_color not in color_pair:
            return False
    return True



# python-mtga-helper
# Copyright 2025 Lubosz Sarnecki <lubosz@gmail.com>
# SPDX-License-Identifier: MIT

import argparse
import json
import logging

from tabulate import tabulate
import numpy as np

from mtga_helper.grading import score_to_grade_string
from mtga_helper.mtg import COLOR_PAIRS, LIMITED_DECK_SIZE, are_card_colors_in_pair, format_color_id_emoji
from mtga_helper.seventeen_lands import has_card_type, count_creatures, get_graded_rankings, print_rankings

logger = logging.getLogger(__name__)

def split_pool_by_color_pair(set_rankings_by_arena_id: dict, pool: list, include_lands=False) -> dict:
    pool_rankings_by_color_pair = {}
    for color_pair in COLOR_PAIRS.keys():
        pool_rankings_by_color_pair[color_pair] = []
        for arena_id in pool:
            if arena_id in set_rankings_by_arena_id:
                ranking = set_rankings_by_arena_id[arena_id]
            else:
                logger.debug(f"Could not find card with arena ID {arena_id}! Hopefully it's a basic land?")
                logger.debug(f"Check scryfall: https://api.scryfall.com/cards/arena/{arena_id}")
                continue

            if not include_lands and has_card_type(ranking, "Land"):
                continue

            if are_card_colors_in_pair(ranking["color"], color_pair):
                pool_rankings_by_color_pair[color_pair].append(ranking)

    return pool_rankings_by_color_pair

def get_top_scores(rankings: list, score_key: str, card_count: int) -> tuple[float, float, float]:
    scores = []
    for ranking in rankings:
        if ranking[score_key]:
            scores.append(ranking[score_key])

    sorted_scores = sorted(scores, reverse=True)
    top_scores = sorted_scores[:card_count]
    worst_of_top = top_scores[-1]
    best_of_top = top_scores[0]
    return float(np.mean(top_scores)), best_of_top, worst_of_top

def color_pair_stats_row(i: int, color_pair: str, score_triple: tuple, rankings: list) -> tuple:
    creature_count, non_creature_count = count_creatures(rankings)
    mean, best, worst = score_triple

    return (
        i + 1,
        f"{format_color_id_emoji(color_pair)} {COLOR_PAIRS[color_pair]}",
        score_to_grade_string(mean),
        mean,
        f"{score_to_grade_string(best)} - {score_to_grade_string(worst)}",
        creature_count,
        non_creature_count,
        len(rankings),
    )

def print_limited_course_info(course: dict, args: argparse.Namespace):
    pool: list = course["CardPool"]
    event_name = course["InternalEventName"]
    logger.info(f"Found limited event {event_name}")

    event_name_split = event_name.split("_")
    assert len(event_name_split) == 3

    set_handle = event_name_split[1].lower()
    logger.info(f"Found event for set handle `{set_handle}`")

    set_rankings_by_arena_id = get_graded_rankings(set_handle, args.data_set, args)

    if args.verbose:
        print(f"== All Rankings for {set_handle.upper()} ==")
        print_rankings(list(set_rankings_by_arena_id.values()))

    target_non_land_count = LIMITED_DECK_SIZE - args.land_count

    # all colors
    pool_rankings = []
    for arena_id in pool:
        if arena_id in set_rankings_by_arena_id:
            pool_rankings.append(set_rankings_by_arena_id[arena_id])
        else:
            logger.debug(f"Could not find card with arena ID {arena_id}! Hopefully it's a basic land?")
            logger.debug(f"Check scryfall: https://api.scryfall.com/cards/arena/{arena_id}")

    print()
    print(f"== {event_name_split[0]} Pool ==")
    print()
    print_rankings(pool_rankings)

    # by color
    pool_rankings_by_color_pair = split_pool_by_color_pair(set_rankings_by_arena_id, pool)
    scores_by_color_pair = {}
    for color_pair, rankings in pool_rankings_by_color_pair.items():
        scores_by_color_pair[color_pair] = get_top_scores(rankings, "ever_drawn_score", target_non_land_count)

    score_by_color_pair_sorted = sorted(scores_by_color_pair.items(), key=lambda item: item[-1], reverse=True)

    # Only print top 1 color pair for draft pools
    if event_name_split[0] == "Sealed":
        print_top_pairs = args.print_top_pairs
    else:
        print_top_pairs = 1

    print()
    print(f"== Top {print_top_pairs} color pairs ==")

    for i, (color_pair, score_triple) in enumerate(score_by_color_pair_sorted):
        if i < print_top_pairs:
            rankings = pool_rankings_by_color_pair[color_pair]

            rank, pair_str, mean_grade, mean_score, grade_range, num_creatures, num_non_creatures, num_non_lands = \
                color_pair_stats_row(i, color_pair, score_triple, rankings)

            table = {
                "Rank": rank,
                f"Top {target_non_land_count} Mean Grade": mean_grade,
                f"Top {target_non_land_count} Mean Score": f"{mean_score:.2f}%",
                f"Top {target_non_land_count} Grade Range": grade_range,
                "Total Creatures": num_creatures,
                "Total Non Creatures": num_non_creatures,
                "Total Non Lands": num_non_lands,
            }
            print()
            print(tabulate(table.items(), headers=(pair_str, "")))
            print()
            print_rankings(rankings, insert_space_at_line=target_non_land_count)
            print()

    table = []
    for i, (color_pair, score_triple) in enumerate(score_by_color_pair_sorted):
        rankings = pool_rankings_by_color_pair[color_pair]
        table.append(color_pair_stats_row(i, color_pair, score_triple, rankings))

    if event_name_split[0] == "Sealed":
        print(f"== Color pair ranking ==")
        print()

        print(tabulate(table, headers=("", "Pair", "Mean", "Score", "Range", "Creatures", "Non Creatures", "Non Lands")))

def premier_draft_pick_cb(draft_status: dict, args):
    event_name = draft_status["EventId"]

    event_name_split = event_name.split("_")
    assert len(event_name_split) == 3
    set_handle = event_name_split[1].lower()

    rankings_by_arena_id = get_graded_rankings(set_handle, args.data_set, args)

    print()
    print(f"== Pack #{draft_status['PackNumber']} Pick #{draft_status['PickNumber']} ==")
    print()

    pack_rankings = []
    for arena_id_str in draft_status["CardsInPack"]:
        arena_id = int(arena_id_str)
        if arena_id in rankings_by_arena_id:
            pack_rankings.append(rankings_by_arena_id[arena_id])
        else:
            logger.debug(f"Could not find card with arena ID {arena_id}! Hopefully it's a basic land?")
            logger.debug(f"Check scryfall: https://api.scryfall.com/cards/arena/{arena_id}")

    if pack_rankings:
        print_rankings(pack_rankings)
    else:
        logger.warning("No known cards in this pack... Is it only a land basic left?")


def bot_draft_pick_cb(event: dict, args):
    target_non_land_count = LIMITED_DECK_SIZE - args.land_count

    draft_status = json.loads(event["Payload"])
    event_name = draft_status["EventName"]

    event_name_split = event_name.split("_")
    assert len(event_name_split) == 3
    set_handle = event_name_split[1].lower()

    rankings_by_arena_id = get_graded_rankings(set_handle, args.data_set, args)

    if args.verbose:
        print(f"== All Rankings for {set_handle.upper()} ==")
        print_rankings(list(rankings_by_arena_id.values()))

    print()
    print(f"== Pack #{draft_status['PackNumber'] + 1} Pick #{draft_status['PickNumber'] + 1} ==")
    print()

    pack_rankings = []
    for arena_id_str in draft_status["DraftPack"]:
        arena_id = int(arena_id_str)
        if arena_id in rankings_by_arena_id:
            pack_rankings.append(rankings_by_arena_id[arena_id])
        else:
            logger.debug(f"Could not find card with arena ID {arena_id}! Hopefully it's a basic land?")
            logger.debug(f"Check scryfall: https://api.scryfall.com/cards/arena/{arena_id}")

    if pack_rankings:
        print_rankings(pack_rankings)
    else:
        logger.warning("No known cards in this pack... Is it only a land basic left?")

    if draft_status["PickedCards"]:
        print()
        print(f"== Pool ==")
        print()

        pool_rankings = []
        for arena_id_str in draft_status["PickedCards"]:
            arena_id = int(arena_id_str)
            if arena_id in rankings_by_arena_id:
                pool_rankings.append(rankings_by_arena_id[arena_id])
            else:
                logger.debug(f"Could not find card with arena ID {arena_id}! Hopefully it's a basic land?")
                logger.debug(f"Check scryfall: https://api.scryfall.com/cards/arena/{arena_id}")

        creature_count, non_creature_count = count_creatures(pool_rankings)
        mean, best, worst = get_top_scores(pool_rankings, "ever_drawn_score", target_non_land_count)

        table = {
            f"Top {target_non_land_count} Mean Grade": score_to_grade_string(mean),
            f"Top {target_non_land_count} Mean Score": f"{mean:.2f}%",
            f"Top {target_non_land_count} Grade Range": f"{score_to_grade_string(best)} - {score_to_grade_string(worst)}",
            "Total Creatures": creature_count,
            "Total Non Creatures": non_creature_count,
            "Total Cards": len(draft_status["PickedCards"]),
        }
        print()
        print(tabulate(table.items()))
        print()

        if pool_rankings:
            print_rankings(pool_rankings, insert_space_at_line=target_non_land_count)
        else:
            logger.warning("Nothing to see here??")


# python-mtga-helper
# Copyright 2025 Lubosz Sarnecki <lubosz@gmail.com>
# SPDX-License-Identifier: MIT

import json
import logging
from datetime import datetime, timezone
from urllib.parse import urlencode

import requests
from tabulate import tabulate
from xdg_base_dirs import xdg_cache_home

from mtga_helper.grading import score_to_grade_string, calculate_grade_scores
from mtga_helper.mtg import format_color_id_emoji, rarity_to_emoji, land_string_to_colors

logger = logging.getLogger(__name__)

APP_NAME = "python-mtga-helper"
CACHE_DIR = xdg_cache_home() / APP_NAME
CACHE_DIR_17LANDS = CACHE_DIR / "17lands"
CACHE_DIR_17LANDS.mkdir(parents=True, exist_ok=True)

def query_17lands(expansion: str, format_name: str):
    params = {
        "expansion": expansion,
        "format": format_name,
        "end_date": datetime.now(timezone.utc).date().isoformat(),
    }
    params_str = urlencode(params)
    cache_file = CACHE_DIR_17LANDS / f"{params_str}.json"

    if not cache_file.is_file():
        logger.info(f"Fetching 17lands data for {params_str}")
        res = requests.get("https://www.17lands.com/card_ratings/data", params=params)
        res.raise_for_status()
        with cache_file.open("w") as f:
            f.write(res.text)
        return res.json()
    else:
        logger.debug(f"Found 17land cache file at {cache_file}")
        with cache_file.open("r") as f:
            return json.loads(f.read())

def get_graded_rankings(set_handle: str, format_name: str, args):
    set_rankings = query_17lands(set_handle, format_name)
    rankings_by_arena_id = {}
    for ranking in set_rankings:

        # Annotate colors on some lands
        if not ranking["color"] and has_card_type(ranking, "Land"):
            for card_type in ranking["types"]:
                ranking["color"] = land_string_to_colors(card_type)

        rankings_by_arena_id[ranking["mtga_id"]] = ranking

    if args.verbose:
        print_rankings_key_histogram(set_rankings)

    return calculate_grade_scores(rankings_by_arena_id, set_rankings)

def has_card_type(ranking: dict, type_name: str) -> bool:
    for card_type in ranking["types"]:
        if type_name in card_type:
            return True
    return False

def count_creatures(rankings: list) -> tuple[int, int]:
    creature_count = 0
    non_creature_count = 0

    for ranking in rankings:
        if has_card_type(ranking, "Creature"):
            creature_count += 1
        else:
            non_creature_count += 1

    return creature_count, non_creature_count

def print_rankings(rankings: list, insert_space_at_line: int = 0):
    table = []
    for ranking in rankings:
        win_rate = 0
        if ranking["ever_drawn_win_rate"]:
            win_rate = ranking["ever_drawn_win_rate"] * 100

        table.append((
            format_color_id_emoji(ranking["color"]),
            rarity_to_emoji(ranking["rarity"]),
            ranking["name"],
            score_to_grade_string(ranking["ever_drawn_score"]),
            f"{win_rate:.2f}",
            " ".join(ranking["types"]),
        ))
    table = sorted(table, key=lambda item: item[-2], reverse=True)

    if insert_space_at_line:
        table_spaced = []
        for i, row in enumerate(table):
            table_spaced.append(row)
            if i == insert_space_at_line - 1:
                table_spaced.append(())
        table = table_spaced

    print(tabulate(table, headers=("", "", "Card", "", "Win %", "Type"), colalign=("right",)))

def print_rankings_key_histogram(rankings):
    keys = [
        "ever_drawn_win_rate",
        "ever_drawn_game_count",
        "drawn_win_rate",
        "win_rate",
    ]

    histogram = {}
    for k in keys:
        histogram[k] = 0

    for card in rankings:
        for k in keys:
            if k in card and card[k]:
                histogram[k] += 1

    histogram["all"] = len(rankings)

    print(tabulate(histogram.items()))